/*======================================================================

    Utility to create an FTL partition in a memory region

    ftl_check.c 1.7 1998/05/10 12:15:28

    The contents of this file are subject to the Mozilla Public
    License Version 1.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The initial developer of the original code is David A. Hinds
    <dhinds@hyper.stanford.edu>.  Portions created by David A. Hinds
    are Copyright (C) 1998 David A. Hinds.  All Rights Reserved.

======================================================================*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <linux/flash.h>
#include <linux/ftl.h>

/*====================================================================*/

static void print_size(u_int s)
{
    if ((s > 0x100000) && ((s % 0x100000) == 0))
	printf("%d mb", s / 0x100000);
    else if ((s > 0x400) && ((s % 0x400) == 0))
	printf("%d kb", s / 0x400);
    else
	printf("%d bytes", s);
}

/*====================================================================*/

static void check_partition(int fd, int verbose, off_t start, size_t size)
{
    flash_info_t info;
    flash_region_info_t *region;
    erase_unit_header_t hdr, hdr2;
    u_int i, j, nbam, *bam;
    int control, data, free, deleted;
    u_int sector_size;

    /* Get partition size, block size */
    if (ioctl(fd, FLASHGETINFO, &info) != 0) {
      perror("get info failed");
      return;
    }
    region = malloc(sizeof(flash_region_info_t)*info.num_regions);
    if (ioctl(fd, FLASHGETREGIONINFO, region) != 0) {
      perror("get region info failed");
      return;
    }
    sector_size = 0;
    for (i=0;i<info.num_regions;i++) {
      if ((start >= region[i].offset) &&
	  (start < region[i].offset + region[i].num_sectors*region[i].sector_size) ) {
	sector_size = region[i].sector_size;
      }
    }

    if (sector_size == 0) {
      printf("Wrong start address!\n");
      return;
    }


    printf("Memory region info:\n");
    printf("  Card offset = 0x%x, region size = ", start);
    print_size(size);
    printf("  Erase block size = ");
    print_size(sector_size);
    printf("\n\n");

    for (i = 0; i < size/sector_size; i++) {
	if (lseek(fd, (i * sector_size), SEEK_SET) == -1) {
	    perror("seek failed");
	    break;
	}
	read(fd, &hdr, sizeof(hdr));
	if ((hdr.FormattedSize > 0) &&
	    (hdr.FormattedSize <= size) &&
	    (hdr.NumEraseUnits > 0) &&
	    (hdr.NumEraseUnits <= size/sector_size))
	    break;
    }
    if (i == size/sector_size) {
	fprintf(stderr, "No valid erase unit headers!\n");
	return;
    }
    
    printf("Partition header:\n");
    printf("  Formatted size = ");
    print_size(hdr.FormattedSize);
    printf(", erase units = %d, transfer units = %d\n",
	   hdr.NumEraseUnits, hdr.NumTransferUnits);
    printf("  Erase unit size = ");
    print_size(1 << hdr.EraseUnitSize);
    printf(", virtual block size = ");
    print_size(1 << hdr.BlockSize);
    printf("\n");
    
    /* Create basic block allocation table for control blocks */
    nbam = (sector_size >> hdr.BlockSize);
    bam = malloc(nbam * sizeof(u_int));

    for (i = 0; i < hdr.NumEraseUnits; i++) {
	if (lseek(fd, (i << hdr.EraseUnitSize), SEEK_SET) == -1) {
	    perror("seek failed");
	    break;
	}
	if (read(fd, &hdr2, sizeof(hdr2)) == -1) {
	    perror("read failed");
	    break;
	}
	printf("\nErase unit %d:\n", i);
	if ((hdr2.FormattedSize != hdr.FormattedSize) ||
	    (hdr2.NumEraseUnits != hdr.NumEraseUnits) ||
	    (hdr2.SerialNumber != hdr.SerialNumber)) {
	    printf("  Erase unit header is corrupt.\n");
	printf("hdr2.FormattedSize(%d) != hdr.FormattedSize(%d)\n",
	       hdr2.FormattedSize, hdr.FormattedSize);
	printf("hdr2.NumEraseUnits(%d) != hdr.NumEraseUnits(%d)\n",
	       (int)hdr2.NumEraseUnits, (int)hdr.NumEraseUnits);
	printf("hdr2.SerialNumber(%d) != hdr.SerialNumber(%d)\n",
	       hdr2.SerialNumber != hdr.SerialNumber);
	printf("hdr2.LogicalEUN=%d\n", hdr2.LogicalEUN);
	} else if (hdr2.LogicalEUN == 0xffff)
	    printf("  Transfer unit, erase count = %d\n", hdr2.EraseCount);
	else {
	    printf("  Logical unit %d, erase count = %d\n",
		   hdr2.LogicalEUN, hdr2.EraseCount);
	    if (lseek(fd, (i << hdr.EraseUnitSize)+hdr.BAMOffset,
		      SEEK_SET) == -1) {
		perror("seek failed");
		break;
	    }
	    if (read(fd, bam, nbam * sizeof(u_int)) == -1) {
		perror("read failed");
		break;
	    }
	    free = deleted = control = data = 0;
	    for (j = 0; j < nbam; j++) {
		if (BLOCK_FREE(bam[j]))
		    free++;
		else if (BLOCK_DELETED(bam[j]))
		    deleted++;
		else switch (BLOCK_TYPE(bam[j])) {
		case BLOCK_CONTROL: control++; break;
		case BLOCK_DATA: data++; break;
		default: break;
		}
	    }
	    printf("  Block allocation: %d control, %d data, %d free,"
		   " %d deleted\n", control, data, free, deleted);
	}
    }
} /* format_partition */

/*====================================================================*/

int main(int argc, char *argv[])
{
    int verbose;
    int optch, errflg, fd;
    struct stat buf;
    off_t start;
    size_t size;
    char *s;
    
    errflg = 0;
    verbose = 0;
    start = 0;
    size = 0;
    while ((optch = getopt(argc, argv, "vs:z:")) != -1) {
	switch (optch) {
	case 'v':
	    verbose = 1; break;
	case 's':
	    start = strtoul(optarg, &s, 0);
	    if ((*s == 'k') || (*s == 'K'))
		start *= 1024;
	    if ((*s == 'm') || (*s == 'M'))
		start *= 1024*1024;
	    break;
	case 'z':
	    size = strtoul(optarg, &s, 0);
	    if ((*s == 'k') || (*s == 'K'))
		size *= 1024;
	    if ((*s == 'm') || (*s == 'M'))
		size *= 1024*1024;
	    break;
	default:
	    errflg = 1; break;
	}
    }
    
    if (size==0) {
	fprintf(stderr, "You must give a size argument\n", argv[0]);
	exit(EXIT_FAILURE);
    }
    
    if (errflg || (optind != argc-1)) {
	fprintf(stderr, "usage: %s [-v] [-s start] -z size device\n", argv[0]);
	exit(EXIT_FAILURE);
    }
    if (stat(argv[optind], &buf) != 0) {
	perror("status check failed");
	exit(EXIT_FAILURE);
    }
    if (!(buf.st_mode & S_IFCHR)) {
	fprintf(stderr, "%s is not a character special device\n",
		argv[optind]);
	exit(EXIT_FAILURE);
    }
    fd = open(argv[optind], O_RDONLY);
    if (fd == -1) {
	perror("open failed");
	exit(EXIT_FAILURE);
    }
    
    check_partition(fd, verbose, start, size);
    close(fd);
    
    exit(EXIT_SUCCESS);
    return 0;
}
